#include <ModbusSerial.h>
#include <Servo.h>  
#include <SoftwareSerial.h>
#include <avr/wdt.h>

#define THRUSTER_LEFT_PIN 11
#define THRUSTER_RIGHT_PIN 10
#define SLAVE_ID 1
#define PWM_LOW 1100
#define PWM_HIGH 1900
#define MODBUS_SERIAL Serial
#define BAUDRATE 115200 
#define TIMEOUT 200
#define LENGTHOF(array) (sizeof(array) / sizeof(*array))
//Battarey parameters
#define R_S 0.01
#define R_L 49900
#define V_REF 2.5
#define V_R1 2000.0
#define V_R2 1000.0
#define VOLTAGE_PIN A1
#define CURRENT_PIN A0
#define ADC_TO_VOLTAGE 5.0 / 1024

// echosounder uart
#define RX_PIN 9
#define TX_PIN 6
#define ECHO_PACKAGE_SIZE 4

//compass
enum CompassState{
  WORKING,
  CALIBRATION
};
const byte READ_DATA_COMMAND[] =  { 0x68, 0x04, 0x00, 0x04, 0x08 };
const byte START_CALIBARTION_COMMAND[] = {0x68, 0x04, 0x00, 0x08, 0x0C};
const byte STOP_CALIBRATION_COMMAND [] = {0x68, 0x04, 0x00, 0x0A, 0x0E};

//Modbus registers
#define LEFT_THRUSTER_REG 1
#define RIGHT_THRUSTER_REG 2 
#define ECHO_REG 3 
#define CURRENT_REG 4
#define VOLTAGE_REG 5
#define BATTERY_REG 6
#define HEADING_REG 7
#define CALIBRATION_REG 8

// ModbusSerial object
ModbusSerial mb (MODBUS_SERIAL, SLAVE_ID, -1);

// Motor objects
Servo thruster_left;
Servo thruster_right;
enum CompassState state;
uint16_t buffer[14];
int index;

SoftwareSerial softSerial =  SoftwareSerial(RX_PIN, TX_PIN);
int echoPackage[ECHO_PACKAGE_SIZE];
int depth = 0;
unsigned long lastPing;

void setup() {
  thruster_left.attach(THRUSTER_LEFT_PIN);
  thruster_right.attach(THRUSTER_RIGHT_PIN);
  softSerial.begin(BAUDRATE);
  pinMode(RX_PIN, INPUT);
  pinMode(TX_PIN, OUTPUT);
  MODBUS_SERIAL.begin (BAUDRATE); // works on all boards but the configuration is 8N1 which is incompatible with the MODBUS standard
  lastPing = millis();
  wdt_disable();
  mb.config (BAUDRATE);
  mb.setAdditionalServerData("Waterstrider low level");
  mb.addHreg (LEFT_THRUSTER_REG, 1500);
  mb.addHreg (RIGHT_THRUSTER_REG, 1500);
  mb.addHreg (ECHO_REG, 0);
  mb.addHreg(BATTERY_REG, 0);
  mb.addHreg(CURRENT_REG, 0);
  mb.addHreg(VOLTAGE_REG, 0);
  mb.addHreg(HEADING_REG, 0);
  mb.addHreg(CALIBRATION_REG, 0);
  state = WORKING;
  pinMode(CURRENT_PIN, INPUT);
  pinMode(VOLTAGE_PIN, INPUT);
  

}
float ConvertFloatValue(uint16_t *buffer)
{
	float wholePart      = (buffer[0] & 0x0F) * 100 + (buffer[1] >> 4) * 10 + (buffer[1] & 0x0F) * 1,
		  fractionalPart = (buffer[2] >> 4) * 10 + (buffer[2] & 0x0F) * 1;

	float result = wholePart + fractionalPart / 100;

	return (buffer[0] >> 4) ? -result : result;
}
void ReadCompassData(){
  if(state == WORKING){
    softSerial.write(READ_DATA_COMMAND, LENGTHOF(READ_DATA_COMMAND));
    while(index < 14){
      if (softSerial.available()) {
          buffer[index] = softSerial.read();
      }
      index ++;
    }
      int heading = ConvertFloatValue(&buffer[10]);
      mb.setHreg(HEADING_REG, heading);
    
    index = 0;
  }
   
}

void CheckCompassState(){
  if(mb.Hreg(CALIBRATION_REG) == 1 && state == WORKING){
    state = CALIBRATION;
    softSerial.write(START_CALIBARTION_COMMAND, LENGTHOF(START_CALIBARTION_COMMAND));
  }
  else if(mb.Hreg(CALIBRATION_REG) == 0 && state == CALIBRATION){
    softSerial.write(STOP_CALIBRATION_COMMAND, LENGTHOF(STOP_CALIBRATION_COMMAND));
    state = WORKING;
  }
  else{
    ReadCompassData();
  }
  
}
void ping(){
   if(millis() - lastPing > TIMEOUT){
     digitalWrite(RX_PIN, LOW);
     delay(2);
     digitalWrite(TX_PIN, HIGH);
     lastPing = millis();
   }
     
}
void ReadDepth(){
    if (softSerial.available() > ECHO_PACKAGE_SIZE) {
          if (softSerial.read() == 0xFF)
          {
            for (int i = 1; i < ECHO_PACKAGE_SIZE; i++) {
              echoPackage[i] = softSerial.read();
            }
            depth = echoPackage[1] * 256 + echoPackage[2];
            mb.setHreg(ECHO_REG, constrain(depth,0,10000));
            softSerial.flush();
          }
    }
}
void BatteryRead(){
  
  int voltage = (analogRead(VOLTAGE_PIN) * ADC_TO_VOLTAGE) / (V_R2 / (V_R1 + V_R2));
  float current = 1000 * (analogRead(CURRENT_PIN) * ADC_TO_VOLTAGE) / ((R_S * R_L) / 1000);
  int battery = (int)(voltage * V_REF);
    mb.setHreg(VOLTAGE_REG, voltage);
    mb.setHreg(CURRENT_REG, current);
    mb.setHreg(BATTERY_REG, battery);
    delay(30);
}


void loop() {
  CheckCompassState();
  BatteryRead();
  mb.task();
  thruster_left.writeMicroseconds(constrain(mb.Hreg(LEFT_THRUSTER_REG), PWM_LOW, PWM_HIGH));
  thruster_right.writeMicroseconds(constrain(mb.Hreg(RIGHT_THRUSTER_REG), PWM_LOW, PWM_HIGH));
  delay(15);
}